﻿package com.ludoko.piratekart 
{
	
	import org.flixel.FlxG;
	import org.flixel.FlxGroup;
	import org.flixel.FlxSprite;
	import org.flixel.FlxState;
	import org.flixel.FlxTileblock;
	import org.flixel.FlxU;
	
	/**
	 * ...
	 * @author Michael Lee
	 */
	public class PlayState extends FlxState
	{
		
		[Embed(source="../../../data/imgs/bg.png")] protected var ImgBG:Class;
		[Embed(source="../../../data/imgs/stair.png")] protected var ImgStair:Class;
		[Embed(source="../../../data/imgs/door.png")] protected var ImgDoor:Class;
		[Embed(source="../../../data/imgs/key.png")] protected var ImgKey:Class;
		[Embed(source="../../../data/imgs/sign.png")] protected var ImgSign:Class;
		
		[Embed(source="../../../data/audio/key.mp3")] protected var SndKey:Class;
		[Embed(source="../../../data/audio/open.mp3")] protected var SndOpen:Class;
		[Embed(source="../../../data/audio/music.mp3")] protected var SndMusic:Class;
		
		private var _stairs:FlxGroup;
		private var _player:Player;
		
		private var _door:FlxSprite;
		private var _key:FlxSprite;
		
		private var _roomWidth:uint, _roomHeight:uint;
		private var _hasKey:Boolean;
		
		override public function create():void
		{
			FlxG.mouse.hide();
			
			_roomWidth = 4800;
			_roomHeight = 4800;
			
			var s:FlxSprite
			s = new FlxSprite(0, 0, ImgBG);
			s.scrollFactor.x = s.scrollFactor.y = 0;
			this.add(s)
			
			_stairs = new FlxGroup();
			
			for (var i:uint = 64; i < _roomWidth; i += 16)
			{
				s = new FlxSprite(i, _roomHeight - 16 - (i - 48), ImgStair);
				s.fixed = true;
				_stairs.add(s);
			}
			
			var tb:FlxTileblock;
			tb = new FlxTileblock(0, _roomHeight - 16, _roomWidth, 16);
			tb.loadGraphic(ImgStair);
			tb.fixed = true;
			_stairs.add(tb);
			tb = new FlxTileblock(0, 0, _roomWidth, 16);
			tb.loadGraphic(ImgStair);
			tb.fixed = true;
			_stairs.add(tb);
			
			tb = new FlxTileblock(0, 0, 16, _roomHeight);
			tb.loadGraphic(ImgStair);
			tb.fixed = true;
			_stairs.add(tb);
			tb = new FlxTileblock(_roomWidth - 16, 0, 16, _roomHeight);
			tb.loadGraphic(ImgStair);
			tb.fixed = true;
			_stairs.add(tb);
			
			this.add(_stairs);
			
			_door = new FlxSprite(16, _roomHeight - 32, ImgDoor);
			this.add(_door);
			
			_key = new FlxSprite(_roomWidth - 32, _roomHeight - (_roomWidth - 48), ImgKey);
			this.add(_key);
			
			this.add(new FlxSprite(32, _roomHeight - 96, ImgSign));
			
			_player = new Player(32, _roomHeight - 32);
			this.add(_player);
			FlxG.follow(_player, 2.5);
			FlxG.followAdjust(0.5, 0.2);
			FlxG.followBounds(0, 0, _roomWidth, _roomHeight);
			
			_hasKey = false;
			
			FlxG.playMusic(SndMusic);
		}
		
		override public function update():void
		{
			super.update();
			
			FlxU.collide(_player, _stairs);
			FlxU.overlap(_player, _door, openDoor);
			FlxU.overlap(_player, _key, getKey);
		}
		
		private function openDoor(ThePlayer:Player, TheDoor:FlxSprite):void
		{
			if (!_hasKey)
				return;
			
			TheDoor.kill();
			FlxG.play(SndOpen);
			FlxG.fade.start(0xff000000, 0.75, endGame);
		}
		
		private function endGame():void
		{
			FlxG.state = new EndState;
		}
		
		private function getKey(ThePlayer:Player, TheKey:FlxSprite):void
		{
			TheKey.kill();
			_hasKey = true;
			FlxG.play(SndKey);
		}
		
	}

}